@@ -41,14 +41,9 @@ class Policy(BaseStochasticFuncType2):
 
     observation_preprocessor : function, optional
 
-        Turns a single observation into a batch of observations that in a form that is convenient
-        for feeding into :code:`func`. If left unspecified, this defaults to:
-
-        .. code:: python
-
-            observation_preprocessor = default_preprocessor(env.observation_space)
-
-        See :func:`coax.utils.default_preprocessor`.
+        Turns a single observation into a batch of observations in a form that is convenient for
+        feeding into :code:`func`. If left unspecified, this defaults to
+        :func:`default_preprocessor(env.observation_space) <coax.utils.default_preprocessor>`.
 
     proba_dist : ProbaDist, optional
 
@@ -111,6 +106,28 @@ class Policy(BaseStochasticFuncType2):
         """
         return super().__call__(s, return_logp=return_logp)
 
+    def mean(self, s):
+        r"""
+
+        Get the mean of the distribution :math:`\pi_\theta(.|s)}`.
+
+        Note that if the actions are discrete, this returns the :attr:`mode` instead.
+
+        Parameters
+        ----------
+        s : state observation
+
+            A single state observation :math:`s`.
+
+        Returns
+        -------
+        a : action
+
+            A single action :math:`a`.
+
+        """
+        return super().mean(s)
+
     def mode(self, s):
         r"""
 
