@@ -21,10 +21,11 @@
 
 import warnings
 
-import jax.numpy as jnp
 import haiku as hk
+import chex
 from gym.spaces import Discrete
 
+from ..utils import is_stochastic
 from ._base import BaseTDLearningQWithTargetPolicy
 
 
@@ -131,24 +132,35 @@ class QLearning(BaseTDLearningQWithTargetPolicy):
         rngs = hk.PRNGSequence(rng)
 
         if isinstance(self.q.action_space, Discrete):
-            # get greedy value directly from q_targ
             params, state = target_params['q_targ'], target_state['q_targ']
             S_next = self.q_targ.observation_preprocessor(next(rngs), transition_batch.S_next)
-            Q_s, _ = self.q_targ.function_type2(params, state, next(rngs), S_next, False)
-            assert Q_s.ndim == 2, f"bad shape: {Q_s.shape}"
-            Q_sa = jnp.max(Q_s, axis=1)
+
+            if is_stochastic(self.q):
+                Q_s = self.q_targ.mean_func_type2(params, state, next(rngs), S_next)
+                Q_s = self.q_targ.proba_dist.postprocess_variate(next(rngs), Q_s, batch_mode=True)
+            else:
+                Q_s, _ = self.q_targ.function_type2(params, state, next(rngs), S_next, False)
+
+            chex.assert_rank(Q_s, 2)
+            assert Q_s.shape[1] == self.q_targ.action_space.n
+
+            # get greedy action as the argmax over q_targ
+            A_next = (Q_s == Q_s.max(axis=1, keepdims=True)).astype(Q_s.dtype)
+            A_next /= A_next.sum(axis=1, keepdims=True)  # there may be ties
 
         else:
-            # get greedy action from pi_targ
+            # get greedy action as the mode of pi_targ
             params, state = target_params['pi_targ'], target_state['pi_targ']
             S_next = self.pi_targ.observation_preprocessor(next(rngs), transition_batch.S_next)
             A_next = self.pi_targ.mode_func(params, state, next(rngs), S_next)
 
-            # evaluate q_targ on greedy action
-            params, state = target_params['q_targ'], target_state['q_targ']
-            S_next = self.q_targ.observation_preprocessor(next(rngs), transition_batch.S_next)
-            Q_sa, _ = self.q_targ.function_type1(params, state, next(rngs), S_next, A_next, False)
+        # evaluate on q_targ
+        params, state = target_params['q_targ'], target_state['q_targ']
+        S_next = self.q_targ.observation_preprocessor(next(rngs), transition_batch.S_next)
 
-        assert Q_sa.ndim == 1, f"bad shape: {Q_sa.shape}"
-        f, f_inv = self.q.value_transform
-        return f(transition_batch.Rn + transition_batch.In * f_inv(Q_sa))
+        if is_stochastic(self.q):
+            return self._get_target_dist_params(params, state, next(rngs), transition_batch, A_next)
+
+        Q_sa_next, _ = self.q.function_type1(params, state, next(rngs), S_next, A_next, False)
+        f, f_inv = self.q.value_transform.transform_func, self.q_targ.value_transform.inverse_func
+        return f(transition_batch.Rn + transition_batch.In * f_inv(Q_sa_next))
